/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MYFITTER_FITTER_HPP
#define MYFITTER_FITTER_HPP

#include "model.hpp"
#include "likelihoodfunction.hpp"

#include <boost/math/special_functions/gamma.hpp>
#include <gsl/gsl_multimin.h>
#include <dvegas/dvegas.h>


namespace myfitter {

// main Fitter class

class Fitter {
public:
    struct FCNParameters {
        Fitter* fitter;
        Model* model;
        const ObservableVector* cvals;
        IndexVector freepar;
        int status;
        int verbose;
    };

    // implementation of Integrand functor for DVegas

    class PValueIntegrand : public HepSource::Integrand {
        Model* fmod_;
        Model* cmod_;
        Fitter* fitter_;
        IndexVector claimedobs_;
        Matrix lam2obs_;
        ObservableVector origin_, lambda0_;
        int ndim0_, ndim1_, ndim_, nvar_;
        double dchisq_;
        double jacobian_;
        int nfailedshots_;
        int verbose_;

        void set_lambda_(const double x[], // adapted integration variables 
                                           // (uses 2 for rsqinner > 0,
                                           // 0 otherwise)
                         const double y[], // normal distributed random
                                           // variables (uses ndim)
                         int ndim,         // number of dimensions to set
                         int offset,       // start at dimension 'offset'
                         double rsqinner,  // square of radius defining
                                           // inner region
                         ObservableVector& lambda, // vector to fill
                         double& jacobian);// this gets multiplied with
                                           // the jacobian of the mapping
    public:
        PValueIntegrand(Model& fmod,       // full model
                        Model& cmod,       // constrained model
                        Fitter& fitter,    // fitter object to fit the models
                        const IndexVector& claimedobs, // claimed observables
                        const Matrix& lam2obs, // conversion from normalised
                                               // to ordinary observables
                        const ObservableVector& origin, // origin in Q-space
                        const ObservableVector& lambda0, // mod[0] best-fit
                                               // point in lambda-space
                        int n0,   // pick radius r in first n0 dimensions
                        int n1,   // pick radius in dimensions n0..n1-1,
                                  // omitting inner region defined by
                                  // r^2 + deltachisq
                        double deltachisq, // delta chi^2 value to beat
                        double jacobian,   // jacobian factor
                        int verbose
                        );

        virtual void operator()(const double x[], const int k[],
                                const double& weight, const double aux[],
                                double f[]);

        // return number of auxiliary dimensions
        int naux() const { return 2*nvar_; }

        // return number of failed shots
        int failed_shots() const { return nfailedshots_; }

        // set number of failed shots to zero
        void reset() { nfailedshots_ = 0; }
    };

private:
    int npar_, nobs_;
    LikelihoodFunction lfunc_;
    Matrix qo_, oq_;
    ObservableVector cobs_, fobs_;

    double orthogonalization_tolerance_;
    const gsl_multimin_fdfminimizer_type* minimizer_;
    int minimizer_verbose_;
    double minimizer_step_width_;
    double minimizer_tolerance_;
    double minimizer_precision_;
    int minimizer_iterations_;
    double minimizer_snap_distance_;
    double inner_region_fraction_;
    double inner_region_power_;

    int dvegas_verbose_;
    int dvegas_nbins_;
    int dvegas_nfirstshots_;
    int dvegas_nshots_;
    int dvegas_niterations_;
    double dvegas_precision_;

    double pvalue_, pvalue_error_;
    bool reached_precision_goal_;
    double failed_shot_ratio_;

    void assign_deriv_(Matrix& m, int icol, const Model& mod, int ipar);
    void set_conversion_matrices_(const Model& fmod, const Model& cmod);
    int local_fit_(Model& model, const ObservableVector& cvals);

public:
    // constructor
    Fitter(int nobs);

    // number of observables
    int nobservables() const { return nobs_; }
    
    // access likelihood function
    LikelihoodFunction& likelihood_function() { return lfunc_; }
    const LikelihoodFunction& likelihood_function() const { return lfunc_; }


    // get/set the the tolerance used for constructing an orthonormal basis
    // from the derivatives of normalized observables with respect to
    // parameters. A (projection of) a derivative vector with respect to
    // parameter i is considered zero if its length is smaller than 
    // orthogonalization_tolerance()*model.scale(i)
    double orthogonalization_tolerance() const
    { return orthogonalization_tolerance_; }
    void orthogonalization_tolerance(double t)
    { orthogonalization_tolerance_ = t; }


    // get/set the gsl minimizer type
    const gsl_multimin_fdfminimizer_type* minimizer() const
    { return minimizer_; }
    void minimizer(const gsl_multimin_fdfminimizer_type* t) { minimizer_ = t; }

    // get/set verbosity level of the gsl minimizer
    int minimizer_verbosity() const { return minimizer_verbose_; }
    void minimizer_verbosity(int n) { minimizer_verbose_ = n; }

    // get/set the initial step width of the gsl minimizer
    double minimizer_step_width() const { return minimizer_step_width_; }
    void minimizer_step_width(double s) { minimizer_step_width_ = s; }

    // get/set the tolerance of the gsl minimizer for 1d minimizations.
    // According to gsl documentation a value of 0.1 is usually sufficient
    double minimizer_tolerance() const { return minimizer_tolerance_; }
    void minimizer_tolerance(double t) { minimizer_tolerance_ = t; }

    // get/set the precision with which the minimizer determines the minimum.
    // The exact meaning of the parameter depends on the minimizer.
    // See the gsl documentation for more information.
    double minimizer_precision() const { return minimizer_precision_; }
    void minimizer_precision(double p) { minimizer_precision_ = p; }

    // get/set the maximum number of iterations for function minimization
    int minimizer_iterations() const { return minimizer_iterations_; }
    void minimizer_iterations(int i) { minimizer_iterations_ = i; }

    // get/set the distance to an upper or lower bound at which a parameter
    // 'snaps' to that bound and the minimization is continued with the
    // parameter fixed at that bound. A parameter i snaps to its upper or
    // lower bound if it's distance to that bound is smaller than
    // minimizer_snap_distance()*scale(i).
    double minimizer_snap_distance() const { return minimizer_snap_distance_; }
    void minimizer_snap_distance(double d) { minimizer_snap_distance_ = d; }
    

    // get/set the initial fraction of points for sampling the inner region
    // in a p-value integral
    double inner_region_fraction() const { return inner_region_fraction_; }
    void inner_region_fraction(double irf) { inner_region_fraction_ = irf; }

    // get/set the power used to construct a mapping for the inner region
    // in a p-value integral
    double inner_region_power() const { return inner_region_power_; }
    void inner_region_power(double irp) { inner_region_power_ = irp; }

    // get/set the verbosity level for p-value integrations
    int dvegas_verbosity() const { return dvegas_verbose_; }
    void dvegas_verbosity(int n) { dvegas_verbose_ = n; }

    // get/set the number of bins for p-value integrations
    int dvegas_nbins() const { return dvegas_nbins_; }
    void dvegas_nbins(int n) { dvegas_nbins_ = n; }

    // get/set the initial number of shots for p-value integrations
    int dvegas_nfirstshots() const { return dvegas_nfirstshots_; }
    void dvegas_nfirstshots(int n) { dvegas_nfirstshots_ = n; }

    // get/set the number of shots per iteration for p-value integrations
    int dvegas_nshots() const { return dvegas_nshots_; }
    void dvegas_nshots(int n) { dvegas_nshots_ = n; }

    // get/set the maximum number iterations for p-value integrations
    int dvegas_niterations() const { return dvegas_niterations_; }
    void dvegas_niterations(int n) { dvegas_niterations_ = n; }

    // get/set the desired relative precision for p-value integrations
    double dvegas_precision() const { return dvegas_precision_; }
    void dvegas_precision(double p) { dvegas_precision_ = p; }

    // calculate chi-square value using predictions of model with current
    // parameters and central observable values cvals. Stores the result in
    // model's chisquare() member.
    int calc(Model& model, const ObservableVector& cvals) {
        int status;
        if((status = model.calc())) return status;
        model.chisquare(lfunc_.calc(model.observables(),cvals)
                        -lfunc_.calc(cvals,cvals));
        return 0;
    }

    // calculate chi-square value using predictions of model with current
    // parameters and the default central values from the likelihood
    // function. Stores the result in model's chisquare() member.
    int calc(Model& model) {
        ObservableVector cvals(nobservables());
        lfunc_.get_central_values(cvals);
        return calc(model, cvals);
    }

    // minimize the likelihood-function with respect to the floating parameters.
    // The data from the last scan() is used to find the best start parameters.
    // Information about the best-fit point is stored in model().
    virtual int global_fit(Model& model, const ObservableVector& cvals);
    int global_fit(Model& model) {
        ObservableVector cvals(nobservables());
        lfunc_.get_central_values(cvals);
        return global_fit(model, cvals);
    }

    // minimize the likelihood-function with respect to the floating parameters.
    // The current parameters of model() are used as start parameters for the
    // minimization. Information about the best-fit point is stored in model().
    virtual int local_fit(Model& model, const ObservableVector& cvals);
    int local_fit(Model& model) {
        ObservableVector cvals(nobservables());
        lfunc_.get_central_values(cvals);
        return local_fit(model, cvals);
    }

    double calc_nested_lrt_pvalue(Model& fullmod, Model& constrainedmod);
    double calc_lrt_pvalue(Model& mod0, Model& mod1, int argc=0, char** argv=0);

    // returns the result of the last p-value integration
    double pvalue() const { return pvalue_; }

    // returns the error of the last p-value integration
    double pvalue_error() const { return pvalue_error_; }

    // checks if the precision goal was reached in the last p-value integration
    bool reached_precision_goal() const { return reached_precision_goal_; }

    // return fraction of failed shots in last p-value integration
    double failed_shot_ratio() const { return failed_shot_ratio_; }
};


}

#endif
