/*
This file is part of myFitter, a C++ library for maximum likelihood fits.
Copyright (C) 2012  Martin Wiebusch

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "likelihoodfunction.hpp"


typedef boost::numeric::ublas::zero_vector<double> ZeroVector;


namespace myfitter {

// class LikelihoodComponent

void LikelihoodComponent::get_central_values(ObservableVector& obs)
{
    assert(int(obs.size()) >= nobservables());
    IndexVector::const_iterator iobs;
    ObservableVector::const_iterator icv;
    for(iobs = claimed_observables().begin(), icv = central_values_.begin();
        iobs != claimed_observables().end(); ++iobs, ++icv)
        obs[*iobs] = *icv;
}


// class LikelihoodFunction

int LikelihoodFunction::add(const LikelihoodComponent& lc)
{
    assert(lc.nobservables() <= nobservables());
    for(IndexVector::const_iterator iobs = lc.claimed_observables().begin();
        iobs != lc.claimed_observables().end(); ++iobs) {
        for(LikelihoodComponentVector::const_iterator ilcv
                = lcv_.begin(); ilcv != lcv_.end(); ++ilcv) {
            if(std::find((*ilcv)->claimed_observables().begin(),
                         (*ilcv)->claimed_observables().end(), *iobs)
               !=(*ilcv)->claimed_observables().end()) return -1;
        }
    }
    lcv_.push_back(lc.copy());
    claimed_observables_.insert(claimed_observables_.end(),
                                lc.claimed_observables().begin(),
                                lc.claimed_observables().end());
    return lcv_.size()-1;
}

void LikelihoodFunction::get_central_values(ObservableVector& meas)
{
    assert(int(meas.size()) == nobservables());
    for(LikelihoodComponentVector::const_iterator ilcv
            = lcv_.begin(); ilcv != lcv_.end(); ++ilcv)
        (*ilcv)->get_central_values(meas);
}

double LikelihoodFunction::calc(const ObservableVector& pred,
                                const ObservableVector& meas)
{
    assert(int(pred.size()) == nobservables());
    assert(int(meas.size()) == nobservables());
    double chi2 = 0.;
    for(LikelihoodComponentVector::const_iterator ilcv
            = lcv_.begin(); ilcv != lcv_.end(); ++ilcv)
        chi2 += (*ilcv)->calc(pred, meas);
    return chi2;
}

int LikelihoodFunction::calc_deriv(const ObservableVector& pred,
                                   const ObservableVector& meas,
                                   ObservableVector& result)
{
    assert(int(pred.size()) == nobservables());
    assert(int(meas.size()) == nobservables());
    assert(int(result.size()) == nobservables());
    result = ZeroVector(nobservables());
    int status;
    for(LikelihoodComponentVector::const_iterator ilcv
            = lcv_.begin(); ilcv != lcv_.end(); ++ilcv)
        if((status=(*ilcv)->calc_deriv(pred, meas, result))) return status;
    return 0;
}

double LikelihoodFunction::calc_nosystematics(const ObservableVector& pred,
                                              const ObservableVector& meas)
{
    assert(int(pred.size()) == nobservables());
    assert(int(meas.size()) == nobservables());
    double chi2 = 0.;
    for(LikelihoodComponentVector::const_iterator ilcv
            = lcv_.begin(); ilcv != lcv_.end(); ++ilcv)
        chi2 += (*ilcv)->calc_nosystematics(pred, meas);
    return chi2;
}

void LikelihoodFunction::get_hessian(Matrix& hess, const ObservableVector& obs)
{
    assert(int(obs.size()) == nobservables());
    hess = boost::numeric::ublas::zero_matrix<double>(nobservables(),
                                                      nobservables());

    for(LikelihoodComponentVector::const_iterator ilcv
            = lcv_.begin(); ilcv != lcv_.end(); ++ilcv)
        (*ilcv)->get_hessian(hess, obs);
}


}
